//standard libraries
#include <stdio.h>
#include <time.h>
#include <stdlib.h>
#include <math.h>
#include <assert.h>

//stantard containers
#include <string>
#include <deque>
#include <list>
#include <vector>
#include <set>
#include <map>
#include <iterator>
//algorithm
#include <algorithm>
//stream
#include <iostream>
#include <fstream>
#include <sstream>
//numerical limits
#include <limits>

//standard name space
using namespace std;

//gsl packages
#include <gsl/gsl_rng.h>		//random number generator for uniform distribution
#include <gsl/gsl_randist.h>	//random number generator for general distributions


//global random number generator
gsl_rng * RNgenerator=gsl_rng_alloc(gsl_rng_default);

//define system log stream
ofstream SysLog("BARD_log.txt",ios::out);

//my own head files
#include "BaseRanker.h"
#include "RankAggregator.h"


/////////////////////////////////////////////////////////////////////
//	***						the main function				 ***  //
/////////////////////////////////////////////////////////////////////
int main(int argc, char *argv[])
/*
arguments:
	[A1] -DataFile: the input data file for 'BaseRankers' (default value is "BaseRankerProfile.txt")
	[A2] -GibbsStepNum: number of Gibbs steps (default value is 1000)
	[A3] -SameGamma: a control parameter for the constraint "gamma_k=c" (default value is False)
	[A4] -Prior4I: hyper-parameter (i.e, p=n_I/n) of the prior distribution for 'Entity Indicators I' (default value is 0.1)
	[A5] -PriorType4Gamma: prior distribution for 'gamma_k', can be either 'exponential distribution' ("Exp") or 'Uniform distribution' ("Unif") (default value is "Exp")
	[A6] -Prior4Gamma_Exp: hyper-parameter of the "Exp" prior for 'gamma_k' (default value is 1)
	[A7] -LowBound4Gamma: low-boundary for 'gamma_k' (default value is 0)
	[A8] -UpBound4Gamma: up-boundary for 'gamma_k' (default value is 10)
	[A9] -RandomSeedBySysTime: reset the random seed baded on system time when 'RandomSeedBySysTime'==1 (default value is 0)
command lines:
	[C1] BARD.exe -DataFile SimuBRP.txt -GibbsStepNum 500 -SameGamma 0 -Prior4I 0.1 -PriorType4Gamma "Exp" -Prior4Gamma_Exp 1 -LowBound4Gamma 0 -UpBound4Gamma 10 -RandomSeedBySysTime 1
function:
	[F1] reset the random seed baded on system time when 'RandomSeedBySysTime'==1
	[F2] load in 'Base Rankers' from 'DataFile' 
	[F3] output the rankings in 'DataFile' into the standard form of 'Base Ranker Profile'
	[F4] rank aggregation
	[F5] output the results into 5 separate files: [O1] "LogLProfile.txt", [O2] "IProfile.txt", [O3] "GammaProfile.txt", [O4] "GammaBar.txt", [O5] "RhoBar.txt"
output:
	[O0] "BARD_Log.txt": system log
	[O1] "LogLProfile.txt": trace plot of 'LogLikelihood'
	[O2] "IProfile.txt": trace plot of 'I'
	[O3] "GammaProfile.txt": trace plot of 'GammaVec'
	[O4] "GammaBar.txt": estimated 'GammaVec' based on the MCMC samples (the first 20% samples in 'GammaProfile' are dropped as the burn-in samples)
	[O5] "RhoBar.txt": estimated 'RhoVec' based on the MCMC samples (the first 20% samples in 'IProfile' are dropped as the burn-in samples)
	[O6] "BARD_BRP.txt": keeps the rankings in 'DataFile' with the standard form of 'Base Ranker Profile'
remark:
	[R1] 'DataFile' supports two formats: 'TopLists' and 'Base Ranker Profile'
*/
{
	//time
	time_t time_start=0, time_end=0, RunningTime=0;
	time(& time_start);

	//set default value for arguments
	string DataFile="BaseRankerProfile.txt";	//[A1]
	unsigned GibbsStepNum=1000;					//[A2]
	bool SameGamma = false;						//[A3]
	double Prior4I=0.1;							//[A4]
	string PriorType4Gamma="Exp";				//[A5]
	double Prior4Gamma_Exp=1;					//[A6]
	double LowBound4Gamma=0;					//[A7]
	double UpBound4Gamma=10;					//[A8]
	bool RandomSeedBySysTime=0;					//[A9]

	//create "ArgName2ArgID"
	map<string,unsigned> ArgName2ArgID;	//map<ArgName,ArgID>
	ArgName2ArgID["DataFile"]=1;
	ArgName2ArgID["GibbsStepNum"]=2;
	ArgName2ArgID["SameGamma"]=3;
	ArgName2ArgID["Prior4I"]=4;
	ArgName2ArgID["PriorType4Gamma"]=5;
	ArgName2ArgID["Prior4Gamma_Exp"]=6;
	ArgName2ArgID["LowBound4Gamma"]=7;
	ArgName2ArgID["UpBound4Gamma"]=8;
	ArgName2ArgID["RandomSeedBySysTime"]=9;

	//specify arguments based on the command line
	string ArgName,ArgValue;
	for(int ArgLoc=1,ArgID=0;ArgLoc<argc;ArgLoc++){
		ArgName=argv[ArgLoc];
		ArgName=ArgName.substr(1,ArgName.size()-1);
		ArgLoc++;
		ArgValue=argv[ArgLoc];

		ArgID=ArgName2ArgID[ArgName];
		if(ArgID==0){	//ignore improper 'Argument'
			cerr<<"Warning: argument '"<<ArgName<<"' is ignored since it cannot be recognized.\n";
			continue;
		}
		switch(ArgID){
			case 1:
				DataFile=ArgValue;
				break;
			case 2:
				GibbsStepNum=atoi(argv[ArgLoc]);
				break;
			case 3:
				SameGamma=atoi(argv[ArgLoc]);
				break;
			case 4:
				Prior4I=atof(argv[ArgLoc]);
				break;
			case 5:
				if(ArgValue=="Unif") PriorType4Gamma=ArgValue;
				break;
			case 6:
				Prior4Gamma_Exp=atof(argv[ArgLoc]);
				break;
			case 7:
				LowBound4Gamma=atof(argv[ArgLoc]);
				break;
			case 8:
				UpBound4Gamma=atof(argv[ArgLoc]);
				break;
			case 9:
				RandomSeedBySysTime=atoi(argv[ArgLoc]);
				break;
		}
	}
	cout  <<"\n------------------------------------------------------------\nInput Arguments:\n------------------------------------------------------------\n";
	cout  <<"[1] DataFile: "<<DataFile<<"\n[2] GibbsStepNum: "<<GibbsStepNum<<"\n[3] SameGamma: "<<SameGamma<<"\n[4] Prior4I: "<<Prior4I<<"\n[5] PriorType4Gamma: "<<PriorType4Gamma<<"\n[6] Prior4Gamma_Exp: "<<Prior4Gamma_Exp<<"\n[7] LowBound4Gamma: "<<LowBound4Gamma<<"\n[8] UpBound4Gamma: "<<UpBound4Gamma<<"\n[9] RandomSeedBySysTime: "<<RandomSeedBySysTime<<endl;
	SysLog<<"\n------------------------------------------------------------\nInput Arguments:\n------------------------------------------------------------\n";
	SysLog<<"[1] DataFile: "<<DataFile<<"\n[2] GibbsStepNum: "<<GibbsStepNum<<"\n[3] SameGamma: "<<SameGamma<<"\n[4] Prior4I: "<<Prior4I<<"\n[5] PriorType4Gamma: "<<PriorType4Gamma<<"\n[6] Prior4Gamma_Exp: "<<Prior4Gamma_Exp<<"\n[7] LowBound4Gamma: "<<LowBound4Gamma<<"\n[8] UpBound4Gamma: "<<UpBound4Gamma<<"\n[9] RandomSeedBySysTime: "<<RandomSeedBySysTime<<endl;

	//[F1] reset random seed based on system time if 'RandomSeedBySysTime'==1
	if(RandomSeedBySysTime){
		int seed = int(time_start);   
		gsl_rng_set(RNgenerator, seed);
	}

	//[F2] load in 'Base Rankers' from 'DataFile' 
	cout  <<"\n------------------------------------------------------------\nLoad in Base Rankers\n------------------------------------------------------------\n";
	SysLog<<"\n------------------------------------------------------------\nLoad in Base Rankers\n------------------------------------------------------------\n";
	RankAggregator RA;
	RA.LoadBaseRanker(DataFile);
	cout<<"m="<<RA.GetM()<<"\nn="<<RA.GetN()<<"\n";

	//[F3] output the rankings in 'DataFile' into the standard form of 'Base Ranker Profile'
	RA.PrintBRP("BARD_BRP.txt");

	//[F4] rank aggregation
	RA.RankAggregation_BARD(Prior4I,PriorType4Gamma,Prior4Gamma_Exp,LowBound4Gamma,UpBound4Gamma,GibbsStepNum,SameGamma);

	//[F5] output the results into 5 separate files: [O1] "LogLProfile.txt", [O2] "IProfile.txt", [O3] "GammaProfile.txt", [O4] "GammaBar.txt", [O5] "RhoBar.txt"
	RA.PrintRunningProfile();
	
	//report running time
	time(& time_end);
	RunningTime = time_end - time_start;
	cout  <<"\nJob finished!\nRunning time: " << RunningTime << " seconds!\n";
	SysLog<<"\nJob finished!\nRunning time: " << RunningTime << " seconds!\n";

	return 0;
}
