////////////////////////////////////////////////////////////////////
//**					Class LogFactorial						**//
////////////////////////////////////////////////////////////////////
/*
decription:
	a component to calculate LogFactorial in an efficient way
function:
	[F1] LogFactorial(n), return the log(n!)
remark:
	[R1] use a deque to keep the possible values, thus avoid repeating computation
	[R2] role in the program: auxiliary class of Class BaseRanker, used in computing the LogLikelihood
*/
class LogFactorial{
private:
	deque<double> memory;

public:
	double operator()(unsigned n){
		if(n<memory.size()) return memory[n];
		memory.push_back(log(double(n))+(* this)(n-1));
		return memory[n];}

	//constructor
	LogFactorial(void){memory.push_back(0);}
	//destructor
	~LogFactorial(){}
};
//a global element for LogFactorial calculation
LogFactorial GlobalLogFactorial;


////////////////////////////////////////////////////////////////////
//**					Class LogInteger						**//
////////////////////////////////////////////////////////////////////
/*
decription:
	a component to calculate Log(n) in an efficient way
function:
	[F1] LogInteger(n), return the log(n)
remark: 
	[R1] use a deque to keep the possible values, thus avoid repeating computation
	[R2] role in the program: auxiliary class of Class BaseRanker, used in computing the LogLikelihood
*/
class LogInteger{
private:
	deque<double> memory;

public:
	//[F1] LogFactorial(n), return the log(n!)
	double operator()(unsigned n){
		if(n<memory.size()) return memory[n];
		memory.push_back(log(double(n))+(* this)(n-1)-(* this)(n-1));
		return memory[n];}

	//constructor
	LogInteger(void){memory.push_back(0);}
	//destructor
	~LogInteger(){}
};
//a global element for LogInteger calculation
LogInteger GlobalLogInteger;


////////////////////////////////////////////////////////////////////
//**					Class BaseRanker						**//
////////////////////////////////////////////////////////////////////
/*
decription:
	data structure to manage data and basic operations of one base ranker
data elements:
	[D1] basic data elements to keep the 'ranker name', ranking data 'tau', quality parameter 'gamma'
	[D2] data elements to efficiently manage entity group indicator 'I' and corresponding sufficient statistics 'tau_{1|0}'
	[D3] data elements related to 'Log-Likelihood'
function:
	[F1] data element get/set functions
	[F2] impute 'tau' to get 'tau_c' based on the current 'I' and 'gamma' via Random Walk Metropolis. Return the Accept Rate of the RWM procedure
	[F3] 'I' related functions
	[F4] 'gamma' related functions
internal functions:
	[F5] get 'tau1_0' from a given 'RankOfRE' based on the current group indicator 'I'
	[F6] calulate the likelihood ratio L(\tau|I,\gamma')/L(\tau|I,\gamma)
remark:
	[R1] the sufficient statistics \tau_{1|0} (i.e, 'Tau1_0') is essential for computing the likelihood, most of the efforts of this class concentrates on managing and updating \tau_{1|0} in an efficient way
	[R2] role in the program: auxiliary class of Class RankAggregator
*/
class BaseRanker{
private:
	//[D1] basic data elements
	string RankerName;					//name of the base ranker
	deque<unsigned> tau;				//ranking of all entities: 'tau[i]' gives the rank of entitiy 'i\in U', tau.size()=|U|, 'tau[i]=0' means that 'tau[i]' is missing
	double gamma;						//the ranker specific power law parameter
	
	//[D2] data elements to efficiently manage 'I' and 'Tau1_0'
	deque<unsigned> tau_c;				//complete version of tau: all missing data in tau are filled up (tau_c=tau if tau is complete)
	set<unsigned> U_M;					//set of entities whose rank is missing
	set<unsigned> R_M;					//set of missing ranks
	set<unsigned> U_R;					//set of relevant entities: keeping information about 'I' for 'tau'	
	map<unsigned,unsigned> RankOfRE;	//map<rank,EntityID>: 'RankOfRE[r]=i' means that 'i\in U_R' and 'tau[i]=r'; interface between 'tau_c' and sufficient statistica 'Tau1_0'

	//[D3] data elements related to 'Log-Likelihood'
	double LogL;						//the log-likelihood based on the current 'tau_c', 'I' and 'gamma'
	double LogB;						//the sufficient statistics 'LogB' for updating 'gamma' based on the current 'tau_c' and 'I'
	double LogL_Ii1;					//the log-likelihood of current 'tau_c' based on the current 'I[-i]' and 'gamma' when I[i]=1
	double LogL_Ii0;					//the log-likelihood of current 'tau_c' based on the current 'I[-i]' and 'gamma' when I[i]=0
	double LogB_Ii1;					//the sufficient statistics 'LogB' for updating 'gamma' based on the current 'tau_c' and'I[-i]' when I[i]=1
	double LogB_Ii0;					//the sufficient statistics 'LogB' for updating 'gamma' based on the current 'tau_c' and'I[-i]' when I[i]=0

	//[F5] get 'tau1_0' from a given 'RankOfRE' based on the current group indicator 'I': Tau1_0 is represented in the form of map< Rank[1|0],set<REID> >. REID="Relavant Entity ID"
	void GetTau1_0(const map<unsigned,unsigned> & RankOfRE,vector<unsigned> & Tau1_0) const{
		//get info
		unsigned n=GetN(), nI=GetNI();
		//clear container
		Tau1_0.clear();	Tau1_0.resize(n-nI+1,0);
		unsigned i=0,Ri=1,Ri1=1,Ri1_0=1;//local variable
		for(map<unsigned,unsigned>::const_iterator REiter=RankOfRE.begin();REiter!=RankOfRE.end();REiter++,Ri1++){
			Ri = (* REiter).first; i = (* REiter).second;	//get info
			Ri1_0 = Ri- Ri1;
			Tau1_0[Ri1_0]++;
		}
	}

	//[F6] functions to calculate 'Log-Likelihood' 
		//[F6-a] calculate 'LogA' and 'LogB' from a given 'Tau1_0', return pair<LogA,LogB>, auxilary function of 'GetLogLB'
	pair<double,double> GetLogAB(const vector<unsigned> & Tau1_0) const{
		double LogA=0,LogB=0;	//clear container
		for(unsigned i=0,L=Tau1_0.size(),RelRank=1,NumOfRE=0;i<L;i++,RelRank++){
			//get info
			NumOfRE=Tau1_0[i];
			//update LogA
			LogA += GlobalLogFactorial(NumOfRE);
			//update LogB
			LogB += ( GlobalLogInteger(RelRank) * NumOfRE);
		}
		pair<double,double> LogAB(LogA,LogB);
		return LogAB;}
		//[F6-b] calculate 'LogC' based on the given 'Gamma' and 'T', auxilary function of 'GetLogLB'
	double GetLogC(double Gamma,double T) const{
		double C=0;
		for(double t=1;t<=T;t++) C += pow(t,-Gamma);
		return log(C);}
		//[F6-c] calculate LogL based on the current 'RankOfRE' and 'gamma', return pair<LogL,LogB>
	pair<double,double> GetLogLB(void) const{
		unsigned n = GetN(), nI = RankOfRE.size();	//get n & nI
		unsigned T = n - nI +1;						//get T = n - nI + 1
		vector<unsigned> Tau1_0;					//get Tau1_0
		GetTau1_0(RankOfRE, Tau1_0);		
		pair<double,double> LogAB =GetLogAB(Tau1_0);//get LogAB
		double LogC = GetLogC(gamma,T);				//get LogC
		double LogLikelihood = GlobalLogFactorial(n-nI) + LogAB.first + double(nI)*LogC + gamma*LogAB.second;
		LogLikelihood =-LogLikelihood;				//get LogLikelihood
		//return LogLB
		pair<double,double> LogLB;
		LogLB.first = LogLikelihood; LogLB.second= LogAB.second;
		return(LogLB);
	}

	//[F2-auxiliary] : replace 'vec[U_BM]' with random samples from 'R_BM' (without replacement)
	void RankImputer(deque<unsigned> & vec,const set<unsigned> & U_BM, const set<unsigned> & R_BM){
		set<unsigned> RankSet = R_BM; unsigned EID = 0, RankValue = 0, RN=0;
		for(set<unsigned>::const_iterator EIDiter=U_BM.begin();EIDiter!=U_BM.end();EIDiter++){
			EID = (* EIDiter);
			RN = gsl_rng_uniform_int( RNgenerator,RankSet.size() );
			set<unsigned>::const_iterator iter=RankSet.begin();
			for(unsigned i=0;i<RN;i++) iter++;
			RankValue = (* iter);
			vec[EID] = RankValue;
			RankSet.erase(RankValue);
		}
	}

public:
	//[F1-a] get access to 'tau_k[i]', return 0 if 'i\notin tau_k' (i.e., R_i is missing for tau_k)
	unsigned operator [](unsigned i) const{return tau[i];}
	//[F1-b] partial list checking function, return 1 if 'tau' is a partial list
	bool IsPartialList(void) const{return !U_M.empty();}

	//[F1-b] get functions for data elements
	string GetName(void) const{return RankerName;}
	deque<unsigned> & GetTau(void){return tau;}
	const deque<unsigned> & GetTau(void) const{return tau;}
	double GetGamma(void) const{return gamma;}
	unsigned GetN(void) const{return tau.size();}
	unsigned GetNI(void) const{return RankOfRE.size();}
	const set<unsigned> & GetUM(void) const{return U_M;}	//get entities whose rank is missing in 'tau'
	const set<unsigned> & GetRM(void) const{return R_M;}	//get ranks that are missing in 'tau'
	double GetLogL(void) const{return LogL;}

	//[F1-d] set functions for data elements
	void SetName(const string & NewValue){RankerName = NewValue;}
	void SetGamma(double NewValue){gamma=NewValue;}
	void SetUR(const set<unsigned> & NewValue){U_R = NewValue;}
	void SetTauC(const deque<unsigned> & NewValue){tau_c=NewValue;}
	void SetRankOfRE(void){	//set 'RankOfRE' based on current 'tau_c' and 'I'
		 RankOfRE.clear();
		for(set<unsigned>::const_iterator EIDiter=U_R.begin();EIDiter!=U_R.end();EIDiter++)
			 RankOfRE[ tau_c[* EIDiter] ] = (* EIDiter);
	}
	unsigned SetMissingRanks(void){//set entities with missing ranks (i.e., U_M) and set of missing ranks (i.e., R_M) based on the current 'tau', return #U_M
		U_M.clear(); R_M.clear();
		for(unsigned i=1,n=tau.size(); i<=n; i++) R_M.insert(i);
		for(unsigned i=0,n=tau.size(); i<n; i++){
			if( tau[i] == 0 ) U_M.insert(i);
			else R_M.erase(tau[i]);
		}
		return U_M.size();}


	//[F2] impute 'tau' to get 'tau_c' based on the current 'I' and 'gamma' via Random Walk Metropolis. Return the Accept Rate of the RWM procedure
	double RankImputer(unsigned RWMStepNum);


	//[F3] 'I' related functions
		//[F3-a] obtain the contribution of base ranker 'tau' to group indicator 'I[i]', i.e., 'DiffLogL'=logP(tau|I_{[I_i=1]},gamma)-logP(tau|I_{[I_i=0]},gamma).
	double EffectOnIi(unsigned i);
		//[F3-b] alter the status of entities: update 'RankOfRE'; reset 'LogL' and 'LogB' from the new 'RankOfRE' based on the current 'I' and 'Gamma'
	unsigned TurnOnEntity(const set<unsigned> & EIDset){
		RankOfRE.clear();	//clear 'RankOfRE' 
		for(set<unsigned>::const_iterator EIDiter=EIDset.begin();EIDiter!=EIDset.end();EIDiter++)
			RankOfRE[ tau_c[* EIDiter] ] = (* EIDiter);	//update 'RankOfRE'
		pair<double,double> LogLB = GetLogLB();			//get LogLB
		LogL = LogLB.first; LogB = LogLB.second;		//set 'LogL' and 'LogB'
		return RankOfRE.size();}
	unsigned TurnOnEntity(unsigned i){
		if( i>= GetN() ){cerr<<"\nError: EntityID overflows!\n";assert(0);}
		RankOfRE[ tau_c[i] ] = i;						//update 'RankOfRE'
		LogL = LogL_Ii1; LogL_Ii1=0; LogL_Ii0=0;		//update 'LogL' and clear 'LogL_Ii1' & 'LogL_Ii0'
		LogB = LogB_Ii1; LogB_Ii1=0; LogB_Ii0=0;		//update 'LogB' and clear 'LogB_Ii1' & 'LogB_Ii0'
		return RankOfRE.size();}
	unsigned TurnOffEntity(unsigned i){
		if( i>= GetN() ){cerr<<"\nError: EntityID overflows!\n";assert(0);}
		RankOfRE.erase( tau_c[i] );						//update 'RankOfRE'
		LogL = LogL_Ii0; LogL_Ii1=0; LogL_Ii0=0;		//update 'LogL' and clear 'LogL_Ii1' & 'LogL_Ii0'
		LogB = LogB_Ii0; LogB_Ii1=0; LogB_Ii0=0;		//update 'LogB' and clear 'LogB_Ii1' & 'LogB_Ii0'
		return RankOfRE.size();}
	

	//[F4] 'gamma' related functions
		//[F4-a] calulate 'DiffLogL(gamma,NewGamma) = log[ L(tau|I,NewGamma)/L(tau|I,gamma) ]'
	double EffectOnGamma(double NewGamma)  const;
		//[F4-b] update 'gamma' with a new value, update 'LogL' accordingly
	void UpdateGamma(double NewGamma){
		//calculate 'DiffLogL'
		double n=GetN(), nI=GetNI(), T=n-nI+1;
		double LogC=GetLogC(gamma,T), NewLogC=GetLogC(NewGamma,T);
		double DiffLogL = nI * (LogC - NewLogC) + LogB * (gamma - NewGamma);
		//update 'LogL' and 'gamma'
		LogL += DiffLogL;
		gamma = NewGamma;}
	
	//constructor
	BaseRanker(){LogL=0;LogB=0;LogL_Ii1=0;LogL_Ii0=0;LogB_Ii1=0;LogB_Ii0=0;gamma=1;}
	//destructor
	~BaseRanker(){}
};


//[F2] impute 'tau' to get 'tau_c' based on the current 'I' and 'gamma' via Random Walk Metropolis. Return the Accept Rate of the RWM procedure
double BaseRanker::RankImputer(unsigned RWMStepNum){
	//get info
		//map<unsigned,unsigned> RankOfRE: relevant entities (i.e., 'U_R') and their ranks in 'tau'
		//set<unsigned> U_M: entities whose rank is missing in 'tau'
		//set<unsigned> R_M: set of missing ranks in 'tau'
	unsigned n = GetN(), nM=U_M.size(); double AR=1; pair<double,double> LogLB, LogLB_new;

	//security check
	if(nM==0) return AR;

	//get relevant entities whose rank is missing (U_RM) and background entities whose rank is missing (U_BM)
	set<unsigned> U_RM, U_BM=U_M, R_RM=R_M, R_BM=R_M;				
	for(set<unsigned>::iterator EIDiter=U_R.begin();EIDiter!=U_R.end();EIDiter++){//scan 'U_R' to set up 'U_RM'
		unsigned EID=(* EIDiter);		//get ID of the Relevant Entity (RE)  
		set<unsigned>::iterator iter=U_M.find(EID);
		if( iter!=U_M.end() ){			//if RE's rank is missing, add it into 'U_RM' and remove it from 'U_BM'
			U_RM.insert(EID); U_BM.erase(EID);}
	}
	//get size of 'U_RM' and 'nBM', respectively
	unsigned nRM=U_RM.size(), nBM=U_BM.size();

	//if 'U_RM' is empty, impute 'R_BM' randomly
	if( nRM == 0 ){
		tau_c = tau;
		RankImputer(tau_c,U_BM,R_BM);	//impute 'tau_c[U_BM]' with random samples (without replacement) from 'R_BM'
		return AR;
	}
	//otherwise, use a two-step procedure to impute 'tau': (1) impute 'R_RM', (2) impute 'R_BM'
	tau_c = tau;					//step 0: initialize 'tau_c' by imputing 'R_RM' randomly
	RankImputer(tau_c,U_RM,R_RM);			//impute 'R_RM' (Rank of Missing Relevant entities) 
	SetRankOfRE();							//set 'RankOfRE' based on the imputed 'tau_c'
	LogLB = GetLogLB();						//calculate the LogL of the current setting

	deque<unsigned> tau_c_old;		//step 1: update 'R_RM' via Random Walk Metroplis
	double LogL_old=0,LogB_old=0,LogL_new=0,MHR=0,RN=0; AR=1;
	for(unsigned t = 1; t < RWMStepNum; t++){
		tau_c_old = tau_c;					//keep previous 'tau_c' in a temporary container
		//LogL_old  = LogL;					//keep previous 'LogL'  in a temporary container
		//LogB_old  = LogB;					//keep previous 'LogB'  in a temporary container

		RankImputer(tau_c,U_RM,R_RM);		//propose a new setting
		SetRankOfRE();						//set 'RankOfRE' based on the new 'tau_c'
		LogLB_new = GetLogLB();				//calculate the LogL of the current setting
										
		MHR = exp(LogLB_new.first - LogLB.first);	//calculate the MH ratio for the MH move
		RN=gsl_rng_uniform(RNgenerator);
		if(RN <= MHR){						//if accept the new setting, update 'LogLB' and 'AR'
			AR++; LogLB = LogLB_new;}
		else tau_c = tau_c_old;				//otherwise, recover the old setting
	}
	
	SetRankOfRE();					//step 2: impute 'R_BM' randomly
	for(map<unsigned,unsigned>::iterator REiter=RankOfRE.begin();REiter!=RankOfRE.end();REiter++)
		R_BM.erase( (* REiter).first );	//scan 'R_R' to filter out the used ranks from 'R_BM'
	RankImputer(tau_c,U_BM,R_BM);
	LogL = LogLB.first; LogB = LogLB.second;//update 'LogL' and 'LogB' for 'tau'

	//update and return 'AR'
	AR/=RWMStepNum;						
	return AR;
}


//[F3-a] obtain the contribution of base ranker 'tau' to group indicator 'I[i]', i.e., 'DiffLogL'=logP(tau|I_{[I_i=1]},gamma)-logP(tau|I_{[I_i=0]},gamma).
double BaseRanker::EffectOnIi(unsigned i)
	//'i': EntityID
	//return: 'DiffLogL'=logP(tau|I_{[I_i=1]},gamma)-logP(tau|I_{[I_i=0]},gamma)
{
	//local variables
	unsigned n = GetN(), nI = 0, Ri=tau_c[i];
	map<unsigned,set<unsigned> > Tau1_0;
	double DiffLogL=0;
	pair<double,double> LogLB;

	//obtain the "LogLikelihood" of the current configuration if it is unknown. This situation happens only at the first step of the 'I' updating procedure
	if(LogL == 0){
		LogLB= GetLogLB();
		LogL = LogLB.first;			//set LogL
		LogB = LogLB.second;		//set LogB
	}

	//get DiffLogL = LogL_Ii1 - LogL_Ii0
	if( RankOfRE.find(Ri) != RankOfRE.end() ){	//if 'I[i]==1' for current 'I'
		LogL_Ii1 = LogL;				//set LogL_Ii1
		LogB_Ii1 = LogB;				//set LogB_Ii1

		RankOfRE.erase(Ri);				//modify 'I' to 'I[i]=0'
		LogLB= GetLogLB();				//get 'LogL' and 'LogB' when 'I[i]=0'
		LogL_Ii0 = LogLB.first;			//set LogL_Ii0
		LogB_Ii0 = LogLB.second;		//set LogB_Ii0

		RankOfRE[Ri]=i;					//recover 'I' to 'I[i]=1'
	}
	else{										//if 'I[i]==0' for current 'I'
		LogL_Ii0 = LogL;				//set LogL_Ii1
		LogB_Ii0 = LogB;				//set LogB_Ii1

		RankOfRE[Ri]=i;					//modify 'I' to 'I[i]=0'
		LogLB= GetLogLB();				//get 'LogL' and 'LogB' when 'I[i]=0'
		LogL_Ii1 = LogLB.first;			//set LogL_Ii0
		LogB_Ii1 = LogLB.second;		//set LogB_Ii0

		RankOfRE.erase(Ri);				//recover 'I' to 'I[i]=1'
	}
	//get and return DiffLogL
	DiffLogL = LogL_Ii1 - LogL_Ii0;
	return DiffLogL;
}

//[F4-a] calulate 'DiffLogL(gamma,NewGamma) = log[ L(tau|I,NewGamma)/L(tau|I,gamma) ]'
double BaseRanker::EffectOnGamma(double NewGamma) const{
	//local variables
	double DiffLogL=0, LogC=0, NewLogC=0;
	unsigned n=GetN(), nI=GetNI(); double T=n-nI+1;

	//calculate the MH ratio (MHR)
	LogC=GetLogC(gamma,T), NewLogC=GetLogC(NewGamma,T);
	DiffLogL = nI * (LogC - NewLogC) + LogB * (gamma - NewGamma);

	//return DiffLogL
	return DiffLogL;
}
