VolumeDistribution <- function(pi,SampleSize)
	## argument: [1] "pi": an n-element vector to present a distribution on {1,...,n}
	#			 [2] "SampleSize": number of entities

	## return:	 [1] "VolDis": empirical distribution of "box volume" when the "box" is generate from independent power law distribution with 'nu' as the parameter (a vector of length="SampleSize")
{
	n <- length(pi)
	LocSpace <- seq(1,n)
	
	VolDis <- rep(0,SampleSize)
	for(i in 1:SampleSize){
		LocPair <- sample(LocSpace, 2, replace = FALSE, prob=pi)
		VolDis[i] <- ( sum(pi[LocPair[1]:LocPair[2]]) )^2
	}
	return(VolDis)
}

### calculate the matrix of coordination coefficients for a given BRP
GetCoordMatrix_BetaDis <- function(BRP)
	## BRP: a 'n' by 'm' matrix for rankings from 'm' base rankers
	
	## return: [1] "AlphaVec4Entity": the fitted 'alpha' paramter for entities
	#		   [2] "BetaVec4Entity": the fitted 'beta' parameter for entities
	#		   [3] "CoordMatrix": the matrix of P-values for 'Coordination Coefficients'
{
	NM <- dim(BRP);
	n <- NM[1];
	m <- NM[2];

	## fit a Beta distribution for each entity
	library(MASS) 			# load package for fitting Beta distribution
	AlphaVec <- rep(0,n)	# define 'AlphaVec'
	BetaVec  <- rep(0,n)	# define 'BetaVec'
	for(i in 1:n){
		TranferedData <- BRP[i,]/(n+1)
		SampleMean<-mean(TranferedData)
		SampleVar <- var(TranferedData)
		Ratio <- SampleMean*(1-SampleMean)/SampleVar-1
		if( Ratio>0 ){
			AlphaVec[i]<-SampleMean*Ratio
			BetaVec[i] <-(1-SampleMean)*Ratio
		}
		else{
			FitBetaDistr <- fitdistr(TranferedData,"beta",start=list(shape1=1,shape2=1))
			MLE_Beta<-FitBetaDistr$estimate
			AlphaVec[i]<-MLE_Beta[1]
			BetaVec[i] <-MLE_Beta[2]
		}
	}

	## get Coordination Coefficient Matrix
	CoordMatrix <- matrix(0,m,m)
	for(i in 1:n){
		# get "pi_Beta"
		Alpha_i <- AlphaVec[i]
		Beta_i  <- BetaVec[i]
		pi_Beta <- dbeta(seq(1,n)/(n+1), Alpha_i, Beta_i); CBeta <- sum(pi_Beta); pi_Beta <- pi_Beta/CBeta;
		# get "VolDis"
		VolDis <- VolumeDistribution(pi_Beta,10000)

		# get pixel-level Pvalue
		for(j1 in 1:m){
			for(j2 in j1:m){
				LocPair <- c(BRP[i,j1],BRP[i,j2])
				Vol <- ( sum(pi_Beta[LocPair[1]:LocPair[2]]) )^2
				Pvalue <- sum(VolDis<Vol)/length(VolDis)
				if(Pvalue==0) Pvalue=1/length(VolDis)
				CoordMatrix[j1,j2] <- CoordMatrix[j1,j2] - log(Pvalue)
				CoordMatrix[j2,j1] <- CoordMatrix[j1,j2]
			}
		}	
	}
	CoordMatrix <- CoordMatrix/n
	## return "rho", "GammaBar" and "LogL"
	return( list("AlphaVec"=AlphaVec, "BetaVec"=BetaVec, "CoordMatrix"=CoordMatrix) )
}
