////////////////////////////////////////////////////////////////////
//**			Class Bayesian Rank Aggeagator					**//
////////////////////////////////////////////////////////////////////
/*
decription:
	data structure to manage a group of base rankers and achieve 'Rank Aggregation'
data elements:
	[D1] basic data elements of 'Base Rankers'
	[D2] data elements to keep learning process profile
function:
	[F1] data element get functions
	[F2] IO functions
	[F3] rank aggregation based on 'Average Rank'
	[F4] rank aggregation based on 'BARD'
remark:
	[R1] the sufficient statistics \tau_{1|0} (i.e, 'Tau1_0') is essential for computing the likelihood, most of the efforts of this class concentrates on managing and updating \tau_{1|0} in an efficient way
	[R2] role in the program: auxiliary class of Class RankAggregator
*/
class RankAggregator{
private:
	//[D1] basic data elements of 'Base Rankers'
	map<string,unsigned> U;		//map<EntityName,EntityID>
	deque<string> EntityVec;	//vector<EntityName>
	deque<BaseRanker> taus;		//keep the m base rankers, \tau_1, ..., \tau_m	
	set<unsigned> U_R;			//vector<GroupIndicator>, I[i]=1 if 'i' is a relevant entity, I[i]=0 if 'i' is a background entity

	//[D2] data elements to keep learning process profile
	deque< set<unsigned> >	IProfile;		//deque< set<U_R> > : IProfile[t] gives the U_R at step t
	deque< double > LogLProfile;			//deque< LogL > : LogLProfile[t]  gives the LogL at step t
	deque< vector<double> > GammaProfile;	//deque< vector<gamma> >: GammaProfile[t][k] give the quality parameter gamma of base ranker 'tau_k' at step t

	//[F2-a1] load in 'BaseRankers' in form of 'BRP', return number of 'BaseRankers' (i.e., 'm')
	unsigned LoadBRP(istream & is);
	//[F2-a2] load in 'BaseRankers' in form of 'TopList', return number of 'BaseRankers' (i.e., 'm')
	unsigned LoadTopList(istream & is);

	//[F4-a]: initialize 'I', 'gamma', 'tau_c', 'LogB' and 'LogL' for 'tau_1', ..., 'tau_m'
	void initialization(double Prior4I,double IniGamma);
	//[F4-b]: impute partial base rankers based on the current parameters
	void GibbsStepForTauC(unsigned RWMStepNum){
		for(unsigned k = 0, m = GetM(); k < m; k++)
			if( taus[k].IsPartialList() ) taus[k].RankImputer(RWMStepNum);}
	//[F4-c]: one Gibbs step for updating 'I'
	void GibbsStepForI(double Prior4I);
	//[F4-d]: one Gibbs step for updating 'Gamma'
	void GibbsStepForGamma(const string & PriorType,double HyperParameter_Exp,double LowBound4Gamma, double UpBound4Gamma, double RWMStepSize,unsigned RWMStepNum,bool SameGamma);
	
public:
	//[F1] data element get functions
	unsigned GetN(void) const{return U.size();}
	unsigned GetM(void) const{return taus.size();}
	unsigned GetNI(void) const{return U_R.size();}
	double GetFullLogL(void) const{
		double FullLogL=0;
		for(unsigned k = 0, m = GetM(); k < m; k++) FullLogL += taus[k].GetLogL();
		return FullLogL;}


	//[F2-a] IO functions: load in 'BaseRankers' form input stream/file, return number of 'BaseRankers' (i.e., 'm')
	unsigned LoadBaseRanker(istream & is){
		//local variables
		char buffer=is.peek(); unsigned m=0;
		if(buffer=='E'){m=LoadBRP(is);return m;}		//if the input file is in form of Base Ranker Profile
		if(buffer=='R'){m=LoadTopList(is);return m;}	//if the input file is in form of Top-d List
		cerr<<"\nBase Ranker Profile has an incorrect format, loading process is terminated!\n";
		return 0;}
	unsigned LoadBaseRanker(const string & InFilePath){
		ifstream ifs(InFilePath.data(),ios::in);
		if(!ifs){cerr<<"File '"<<InFilePath<<"' can not be opened!"<<endl;assert(0);}
		return LoadBaseRanker(ifs);}
	
	//[F2-b] IO functions: output 'BRP' or 'RunningProfile'
	ostream & PrintBRP(ostream & os);
	ostream & PrintBRP(const string & OutFilePath){
		ofstream ofs(OutFilePath.data(),ios::out);
		if(!ofs){cerr<<"File '"<<OutFilePath<<"' can not be opened!"<<endl;assert(0);}
		return PrintBRP(ofs);}
	void PrintRunningProfile(void);

	
	//[F3] rank aggregation based on 'Average Rank'
	void RankAggregation_AveRank(vector<double> & AveRankVec,map<double,set<unsigned> > & AggregatedRank) const
		//vector<double> AveRankVec: vector<AveRank>, AveRankVec[i] gives the AveRank of entity 'i\in U'
		//map<double,set<unsigned> > AggregatedRank: map< Rank,set<EntityID> >, sorted list of enetities
	{
		unsigned n=EntityVec.size();double rank=0;
		AveRankVec.clear();AveRankVec.resize(n,0);
		vector<double> NumOfEffectiveRanker(n,0);
		for(unsigned i=0;i<n;i++){
			for(unsigned k=0,m=taus.size();k<m;k++){
				rank=taus[k][i];
				AveRankVec[i]+=rank;
				if(rank>0) NumOfEffectiveRanker[i]++;}
			AveRankVec[i]/=NumOfEffectiveRanker[i];
			AggregatedRank[ AveRankVec[i] ].insert(i);
		}
	}

	//[F4] rank aggregation based on 'BARD'
	void RankAggregation_BARD(double Prior4I,const string & PriorType4Gamma,double Prior4Gamma_Exp,double LowBound4Gamma,double UpBound4Gamma,unsigned GibbsStepNum,bool SameGamma){
		//local variables
		double GammaRWMStepSize=0.2, FullLogL=0; unsigned m=GetM(), n=U.size(), GammaRWMStepNum=50, TauCRWMStepNum=100;
		vector<double> GammaVec(m,0);

		//initialization: initialize 'I' and 'gamma' for 'tau_1', ..., 'tau_m'.
		double IniGamma=1;
		initialization(Prior4I,IniGamma);
		//Gibbs sampler for [I,Gamma,TauCs]
		for(unsigned StepID=1;StepID<=GibbsStepNum;StepID++){
			//output
			SysLog<<"\n------------------------------------------------------------\nStep "<<StepID<<"\n------------------------------------------------------------\n";

			//one Gibbs step
			GibbsStepForTauC(TauCRWMStepNum);	//update 'TauCs' given [I,Gamma]
			GibbsStepForI(Prior4I);				//update 'I' given [Gamma, TauCs]
												//update 'Gamma' given [I,TauCs] with the uniform prior
			GibbsStepForGamma(PriorType4Gamma,Prior4Gamma_Exp,LowBound4Gamma,UpBound4Gamma,GammaRWMStepSize,GammaRWMStepNum,SameGamma);
			FullLogL = GetFullLogL();			//get the FullLogL of the updated model
			
			//output
			SysLog<<"\ngamma:\n";
			for(unsigned k = 0; k < m; k++){
				GammaVec[k]=taus[k].GetGamma();
				SysLog<<'['<<k<<"]\t"<<GammaVec[k]<<endl;
			}
			SysLog<<"\nRelevant Entities:\n";
			for(set<unsigned>::const_iterator EIDiter=U_R.begin(); EIDiter!=U_R.end(); EIDiter++)
				SysLog<<(* EIDiter)<<endl;
			cout<<'['<<StepID<<"]\tnI="<<U_R.size()<<"\tLogL="<<FullLogL<<endl;

			//update profiles of Gibbs Sampler
			LogLProfile.push_back(FullLogL);
			IProfile.push_back(U_R);
			GammaProfile.push_back(GammaVec);
		}
	}

	//constructor and destructor
	RankAggregator(void){}
	~ RankAggregator(){}
};


//[F2-a1] load in 'BaseRankers' form input stream, return number of 'BaseRankers'(i.e., 'm'). The format is defined in "BaseRankerProfileAlt.txt"
unsigned RankAggregator::LoadBRP(istream & is){
	//local variables
	char buffer = 'a'; string bufflen, RankerName, EntityName; unsigned m = 0, EntityID = 0, k = 0, rank = 0;
	//get 'm'
	getline(is,bufflen);//read in the head line from 'is'
	SysLog<<bufflen<<endl;
	for(unsigned i=0,L=bufflen.size();i<L;i++)
		if(bufflen[i]=='\t') m++;
	//create 'LabelForBR', LabelForBR[k]=1 if loading process of 'tau_k' is not finished, and 0 otherwise
	vector<bool> LabelForBR(m,1);
	//create tau_1, ..., tau_m
	taus.clear();
	BaseRanker NewBR;
	stringstream ss(bufflen,ios::in);
	getline(ss,RankerName,'\t');	//read through the first entry "Rank"
	for(unsigned k = 0; k < m; k++){
		getline(ss,RankerName,'\t');//read in "RankerName"
		NewBR.SetName(RankerName);	//set "RankerName"
		taus.push_back(NewBR);		//update "taus"
	}
	
	//load in BaseRankers
	for(unsigned LineID = 1; !is.eof(); LineID++){
		//read in the current line from 'is'
		getline(is,bufflen);
		SysLog<<bufflen<<endl;
		if( bufflen.empty() ) break;
		stringstream ss(bufflen,ios::in);
		//get 'EntityName'
		EntityName.clear();getline(ss,EntityName,'\t');
		if(EntityName.empty()){cerr<<"\nEntity "<<LineID-1<<" does not have a correct name, loading process is terminated!\n";assert(0);}
		//get 'EntityID', and update 'U' and 'EntityVec'
		map<string,unsigned>::const_iterator EntityIter = U.find(EntityName);
		if(EntityIter == U.end()){
			EntityID=LineID-1; U[EntityName] = EntityID;
			EntityVec.push_back(EntityName);}
		else{cerr<<"\nTwo entities (No."<< (* EntityIter).second<<"and No. "<<LineID-1<<") have the same name '"<<EntityName<<"', loading process is terminated!\n";assert(0);}
		for(k = 0; !ss.eof(); k++){
			buffer = ss.peek();
			if(buffer == '\t' || buffer == EOF){ss.get();continue;}
			//update 'tau_k'
			ss >> rank;//get 'tau_k[i]', i = LineID-1
			taus[k].GetTau().push_back(rank);
			buffer = ss.get();
		}
		if(k != m){cerr<<"\nError: fail to read in line "<<LineID<<" of BaseRanker profile!\n";assert(0);}
	}
	return m;
}
//[F2-a2] load in 'BaseRankers' form input stream, return number of 'BaseRankers'(i.e., 'm'). The format is defined in "BaseRankerProfile.txt"
unsigned RankAggregator::LoadTopList(istream & is){
	//local variables
	char buffer = 'a'; string bufflen, RankerName, EntityName; unsigned m = 0, EntityID = 0, EntityRank=1, ColumNum=0;
	//get 'm'
	getline(is,bufflen);//read in the head line from 'is'
	cout<<bufflen<<endl;
	for(unsigned i=0,L=bufflen.size();i<L;i++)
		if(bufflen[i]=='\t') m++;
	//create 'LabelForBR', LabelForBR[k]=1 if loading process of 'tau_k' is not finished, and 0 otherwise
	vector<bool> LabelForBR(m,1);
	//create tau_1, ..., tau_m
	taus.clear();
	BaseRanker NewBR;
	stringstream ss(bufflen,ios::in);
	getline(ss,RankerName,'\t');	//read through the first entry "Rank"
	for(unsigned k = 0; k < m; k++){
		getline(ss,RankerName,'\t');//read in "RankerName"
		NewBR.SetName(RankerName);	//set "RankerName"
		taus.push_back(NewBR);		//update "taus"
	}
	//load in BaseRankers
	for(unsigned LineID=2;!is.eof();LineID++,EntityRank++){
		getline(is,bufflen);//read in the next line from 'is'
		cout<<bufflen<<endl;
		if( bufflen.empty() ) break;
		stringstream ss(bufflen,ios::in);
		//read through 'Rank'
		getline(ss,EntityName,'\t');
		for(ColumNum = 0; !ss.eof(); ColumNum++){
			//get 'EntityName'
			EntityName.clear();getline(ss,EntityName,'\t');
			if(EntityName.empty()){LabelForBR[ColumNum] = 0; continue;}
			//get 'EntityID', and update 'U' and 'EntityVec'
			map<string,unsigned>::const_iterator EntityIter=U.find(EntityName);
			if(EntityIter==U.end()){
				EntityID=U.size();U[EntityName]=EntityID;
				EntityVec.push_back(EntityName);
				for(unsigned k = 0; k < m; k++) taus[k].GetTau().push_back(0);
			}
			else EntityID=(* EntityIter).second;
			//update 'tau_k'
			deque<unsigned> & RankProfile = taus[ColumNum].GetTau();
			if(LabelForBR[ColumNum]){
				if(EntityID < RankProfile.size() ) RankProfile[EntityID] = EntityRank;
				else RankProfile.push_back(EntityRank);
			}
			else{cerr<<"\nError: the profile of BaseRanker "<<(ColumNum + 1)<<" is not correct at line "<<LineID<<"!\n";assert(0);}
		}
		if(ColumNum != m){cerr<<"\nError: fail to read in line "<<LineID<<" of BaseRanker profile!\n";assert(0);}
	}
	return m;
}

//[F2-b1] data printing functions
ostream & RankAggregator::PrintBRP(ostream & os){
	//local variables
	unsigned m=taus.size(), EntityID = 0; string EntityName;
	//print head line
	os << "Entity";
	for(unsigned k=0; k < m; k++) os << '\t' << taus[k].GetName();
	os << endl;
	//print BRP
	for(map<string,unsigned>::const_iterator EntityIter = U.begin(); EntityIter != U.end(); EntityIter++){
		EntityName = (* EntityIter).first; EntityID = (* EntityIter).second;
		os << EntityName;
		for(unsigned k=0; k < m; k++) os << "\t" << taus[k][EntityID];
		os << endl;
	}
	return os;
}
//[F2-b2] learning process profile printing function
void RankAggregator::PrintRunningProfile(void){
	//create output files 
	ofstream ofs_LogL("LogLProfile.txt",ios::out);	//trace plot of LogL
	ofstream ofs_I("IProfile.txt",ios::out);		//trace plot of I
	ofstream ofs_Gamma("GammaProfile.txt",ios::out);//trace plot of GammaVec
	ofstream ofs_GammaBar("GammaBar.txt",ios::out);	//estimated GammaVec based on the MCMC samples (the first 20% samples in GammaProfile are dropped as the burn-in samples)
	ofstream ofs_RhoBar("RhoBar.txt",ios::out);		//estimated RhoVec based on the MCMC samples (the first 20% samples in IProfile are dropped as the burn-in samples)

	//initialize GammaBar & RhoBar
	unsigned m=GetM(),n=GetN();
	vector<double> GammaBar(m,0),RhoBar(n,0);

	//print head lines of output files
	ofs_LogL<<"Step\tLogL\n";
	ofs_I<<"Step"; for(unsigned i=0;i<n;i++) ofs_I<<'\t'<<EntityVec[i]; ofs_I<<endl;
	ofs_Gamma<<"Step"; for(unsigned k=0;k<m;k++) ofs_Gamma<<'\t'<<taus[k].GetName(); ofs_Gamma<<endl;
	ofs_GammaBar<<"RID\tRName\tGammaBar\n";
	ofs_RhoBar<<"EID\tEName\tRhoBar\n";

	//output LogL, GammaVec and I; prepare GammaBar and RhoBar along the printing process
	double StepNum=LogLProfile.size(), BurnIn=0.2*StepNum, EffetiveStep=0;
	for(unsigned t=0;t<StepNum;t++){
		//LogL
		ofs_LogL<<t<<"\t"<<LogLProfile[t]<<endl;
		//GammaVec
		ofs_Gamma<<t; 
		for(unsigned k=0;k<m;k++){
			ofs_Gamma<<'\t'<<GammaProfile[t][k];
			if(t>BurnIn) GammaBar[k] += GammaProfile[t][k];
		}
		ofs_Gamma<<endl;
		//I
		ofs_I<<t;
		for(unsigned i=0,I_i=0;i<n;i++){
			I_i=IProfile[t].find(i)==IProfile[t].end()?0:1;
			ofs_I<<'\t'<<I_i;
			if(t>BurnIn) RhoBar[i] += I_i;
		}
		ofs_I<<endl;
		//update EffetiveStep
		if(t>BurnIn) EffetiveStep++;
	}

	//output GammaBar and RhoBar
	for(unsigned k=0;k<m;k++){
		GammaBar[k]/=EffetiveStep;
		ofs_GammaBar<< k+1 <<"\t" << taus[k].GetName() << '\t' << GammaBar[k] <<endl;}
	for(unsigned i=0;i<n;i++){
		RhoBar[i]  /=EffetiveStep;
		ofs_RhoBar  << i+1 <<"\t" << EntityVec[i]	   << '\t' << RhoBar[i]   <<endl;}
}


//[F4-a] initialize 'I', 'gamma', 'tau_c', 'LogB' and 'LogL' for 'tau_1', ..., 'tau_m'
void RankAggregator::initialization(double Prior4I,double IniGamma){
	//get intital value for 'gamma'
	double gamma=IniGamma, FullLogL=0;
	//set initial value for 'I' based on AveRank
	unsigned m = GetM(), n = GetN(), EntityID = 0;
	unsigned SizeOfU_R = unsigned( Prior4I * double(n) );
	U_R.clear();
	vector<double> AveRankVec;					//vector<AveRank>, AveRankVec[i] gives the AveRank of entity 'i\in U'
	map<double,set<unsigned> > AggregatedRank;	//map< Rank,set<EntityID> >, sorted list of enetities
	RankAggregation_AveRank(AveRankVec,AggregatedRank);
	//for(map<double,set<unsigned> >::const_reverse_iterator RankIter=AggregatedRank.rbegin();RankIter!=AggregatedRank.rend();RankIter++){
	for(map<double,set<unsigned> >::const_iterator RankIter=AggregatedRank.begin();RankIter!=AggregatedRank.end();RankIter++){
		const set<unsigned> & EntitySet=(* RankIter).second;
		for(set<unsigned>::const_iterator EntityIter=EntitySet.begin();EntityIter!=EntitySet.end();EntityIter++){
			if(U_R.size() >= SizeOfU_R) break;
			EntityID = (* EntityIter);
			U_R.insert(EntityID);
		}
	}
	//intialize 'I', 'gamma', 'tau_c', 'LogB' and 'LogL' for 'tau_1', ..., 'tau_m'
	for(unsigned k = 0, RWMStepNum = 1; k < m; k++){
		BaseRanker & tau=taus[k];		//get infor
		tau.SetMissingRanks();			//set 'U_M' and 'R_M'
		tau.SetUR(U_R);					//intialize 'U_R', i.e. 'I'
		tau.SetGamma(gamma);			//intialize 'gamma'
										//set 'tau_c'
		if( tau.IsPartialList() ) tau.RankImputer(RWMStepNum);
		else tau.SetTauC( tau.GetTau() );
		tau.TurnOnEntity(U_R);			//set 'LogB' and 'LogL'
		FullLogL += taus[k].GetLogL();	//update FullLogL
	}

	//initialize profiles of Gibbs Sampler at step 0
	LogLProfile.push_back(FullLogL);
	IProfile.push_back(U_R);
	vector<double> GammaVec(m,gamma);
	GammaProfile.push_back(GammaVec);
}


//[F4-c] one Gibbs step for 'I' given 'Gamma' and 'TauCs'
void RankAggregator::GibbsStepForI(double Prior4I){
	//local variables
	unsigned n = GetN(), m = GetM(), nI = GetNI(), nI0 = 0; double SqrtM=sqrt( double(m) );
	double DiffLogL_TauK = 0, DiffLogL = 0, FullLogL = 0, PI_penalty = 0, q_i = 0, r_i = 0, RN = 0;
	bool I_i = 0;
	//scan through all entities to update 'I'
	SysLog<<"\n---------------------------\nScan through all entities to update 'I'\n---------------------------\nEID\tI\tn_I\tP(I_i=1)\n";
	for(unsigned i = 0; i < n; i++){
		//calculate 'q_i' and 'r_i'
		DiffLogL = 0;
		for(unsigned k = 0; k < m; k++){
			DiffLogL_TauK = taus[k].EffectOnIi(i);
			DiffLogL += DiffLogL_TauK;
		}
		if( DiffLogL > 100 ) r_i = 1;
		else{
			//get nI0 = sum(I) - I_i
			if( U_R.find(i) == U_R.end() ) nI0 = GetNI();
			else nI0 = GetNI() - 1;
			//get PI_penalty
			PI_penalty = (n*Prior4I - nI0 - 0.5) * SqrtM;
			//get r_i = P(I_i=1|I[-i],gamma)
			q_i = exp(PI_penalty + DiffLogL);
			r_i = q_i/(q_i + 1);
		}
		//get NewValue for I[i]
		RN = gsl_rng_uniform(RNgenerator);
		if(RN <= r_i) I_i = 1;
		else I_i = 0;

		//update 'I' 
		if(I_i == 1){
			U_R.insert(i);
			for(unsigned k=0;k<m;k++) taus[k].TurnOnEntity(i);}
		else{
			U_R.erase(i);
			for(unsigned k=0;k<m;k++) taus[k].TurnOffEntity(i);}
		//output
		if(I_i == 1) SysLog<<'['<<i<<"]\tOn\t"<<taus[0].GetNI()<<'\t'<<r_i<<endl;
		else SysLog<<'['<<i<<"]\tOff\t"<<taus[0].GetNI()<<'\t'<<r_i<<endl;
	}
}

//[F4-d] one Gibbs step for 'Gamma' given 'I' and 'TauCs'. If 'SameGamma'=TRUE, sample gamma_k under the constrain that gamma_1=...=gamma_m; otherwise, sample gamma_k independently
void RankAggregator::GibbsStepForGamma(const string & PriorType,double HyperParameter,double LowBound,double UpBound,double RWMStepSize,unsigned RWMStepNum,bool SameGamma)
{
	//local variables
	double gamma=0, OldGamma=0, NewGamma=0, RN=0, DiffLogL =0, MHR=0, AR=0;
	double Exp_Mean=HyperParameter;		//hyper-parameter of 'gamma' (i.e., '\lambda') when 'PriorType'="Exp"

	//'b'=HyperParameter when 'PriorType'="Unif"
	if(PriorType=="Unif") UpBound=HyperParameter;	

	//If 'SameGamma'=TRUE, sample the common gamma
	if( SameGamma == true ){
		//If 'SameGamma'=TRUE, sample gamma_k under the constrain that gamma_1=...=gamma_m
		SysLog<<"\n---------------------------\nScan through all base rankers to update the common 'gamma'\n---------------------------\nOldGamma\tNewGamma\tAcceptRate\n";
		//update common 'gamma' via RWM
		for(unsigned i=0; i<RWMStepNum;i++){
			//propose NewGamma via Gaussian random walk
			gamma=taus[0].GetGamma();
			RN = gsl_ran_gaussian (RNgenerator,RWMStepSize);
			NewGamma = gamma + RN;

			//relect the proposed value between the boundaries
			while( NewGamma < LowBound || NewGamma > UpBound ){
				if( NewGamma < LowBound ) NewGamma = 2 * LowBound - NewGamma; 
				if( NewGamma > UpBound  ) NewGamma = 2 * UpBound  - NewGamma;
			}

			//initialize 'DiffLogL'
			if(PriorType=="Unif") DiffLogL = 0;										//initialize 'DiffLogL' with 0 when 'PriorType'=="Unif"
			if(PriorType=="Exp") DiffLogL = -Exp_Mean*(NewGamma - gamma);	//initialize 'DiffLogL' with the prior when 'PriorType'=="Exp"
			//calculate 'DiffLogL'
			for(unsigned k = 0, m = GetM(); k < m; k++) 
				DiffLogL += taus[k].EffectOnGamma(NewGamma);
			//calculate MH ratio (MHR)
			MHR = exp(DiffLogL);

			// MH move
			RN = gsl_rng_uniform(RNgenerator);
			if(RN <= MHR){
				AR++;	//update AR
				for(unsigned k = 0, m = GetM(); k < m; k++)
					taus[k].UpdateGamma(NewGamma);	//accept NewGamma and update gamma with it
			}
		}
		//output 'gamma', 'NewGamma' and 'AR'
		AR/=double(RWMStepNum);
		SysLog<<OldGamma<<'\t'<<NewGamma<<'\t'<<AR<<endl;

		return;
	}

	//If 'SameGamma'=FALSE, sample gamma_k independently
	SysLog<<"\n---------------------------\nScan through all base rankers to update 'gamma'\n---------------------------\nBaseRanker\tOldGamma\tNewGamma\tAcceptRate\n";
	for(unsigned k = 0, m = GetM(); k < m; k++){
		OldGamma=gamma=taus[k].GetGamma();	//get OldGamma
		AR = 0;								//clear up AR
		for(unsigned i=0; i<RWMStepNum;i++){
			//propose NewGamma via Gaussian random walk
			gamma=taus[k].GetGamma();
			RN = gsl_ran_gaussian (RNgenerator,RWMStepSize);
			NewGamma = gamma + RN;

			//relect the proposed value between the boundaries
			while( NewGamma < LowBound || NewGamma > UpBound ){
				if( NewGamma < LowBound ) NewGamma = 2 * LowBound - NewGamma; 
				if( NewGamma > UpBound  ) NewGamma = 2 * UpBound  - NewGamma;
			}

			//initialize 'DiffLogL'
			if(PriorType=="Unif") DiffLogL = 0;								//initialize 'DiffLogL' with 0 when 'PriorType'=="Unif"
			if(PriorType=="Exp") DiffLogL = -Exp_Mean*(NewGamma - gamma);	//initialize 'DiffLogL' with the prior when 'PriorType'=="Exp"
			//calculate 'DiffLogL'
			DiffLogL += taus[k].EffectOnGamma(NewGamma);
			//calculate MH ratio (MHR)
			MHR = exp(DiffLogL);

			// MH move
			RN = gsl_rng_uniform(RNgenerator);
			if(RN <= MHR){
				AR++;							//update AR
				taus[k].UpdateGamma(NewGamma);	//accept NewGamma and update gamma with it
			}
		}
		//output 'OldGamma', 'NewGamma' and 'AR'
		AR/=double(RWMStepNum);
		SysLog<<'['<<k<<"]\t"<<OldGamma<<'\t'<<NewGamma<<'\t'<<AR<<endl;
	}
}
